/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.protocol.sync;

import org.teleal.cling.binding.xml.DescriptorBindingException;
import org.teleal.cling.binding.xml.DeviceDescriptorBinder;
import org.teleal.cling.binding.xml.ServiceDescriptorBinder;
import org.teleal.cling.model.Resource;
import org.teleal.cling.model.meta.Icon;
import org.teleal.cling.model.meta.LocalDevice;
import org.teleal.cling.model.message.StreamRequestMessage;
import org.teleal.cling.model.message.StreamResponseMessage;
import org.teleal.cling.model.message.UpnpResponse;
import org.teleal.cling.model.message.header.ContentTypeHeader;
import org.teleal.cling.UpnpService;
import org.teleal.cling.model.meta.LocalService;
import org.teleal.cling.protocol.ReceivingSync;
import org.teleal.common.util.Exceptions;

import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;


public class ReceivingRetrieval extends ReceivingSync<StreamRequestMessage, StreamResponseMessage> {

    private static Logger log = Logger.getLogger(ReceivingRetrieval.class.getName());

    public ReceivingRetrieval(UpnpService upnpService, StreamRequestMessage inputMessage) {
        super(upnpService, inputMessage);
    }

    protected StreamResponseMessage executeSync() {

        if (!getInputMessage().hasHostHeader()) {
            log.fine("Ignoring message, missing HOST header: " + getInputMessage());
            return new StreamResponseMessage(new UpnpResponse(UpnpResponse.Status.PRECONDITION_FAILED));
        }

        URI requestedURI = getInputMessage().getOperation().getURI();

        Resource foundResource = getUpnpService().getRegistry().getResource(requestedURI);

        if (foundResource == null) {
            log.fine("No local resource found: " + getInputMessage());
            return null;
        }

        return createResponse(requestedURI, foundResource);
    }

    protected StreamResponseMessage createResponse(URI requestedURI, Resource resource) {
        try {
            switch (resource.getType()) {

                case DEVICE_DESCRIPTOR:

                    log.fine("Found local device matching relative request URI: " + requestedURI);
                    LocalDevice device = (LocalDevice) resource.getModel();

                    DeviceDescriptorBinder deviceDescriptorBinder =
                            getUpnpService().getConfiguration().getDeviceDescriptorBinderUDA10();
                    String deviceDescriptor = deviceDescriptorBinder.generate(
                            device,
                            getUpnpService().getConfiguration().getNamespace()
                    );
                    return new StreamResponseMessage(
                            deviceDescriptor,
                            new ContentTypeHeader(ContentTypeHeader.DEFAULT_CONTENT_TYPE)
                    );

                case SERVICE_DESCRIPTOR:

                    log.fine("Found local service matching relative request URI: " + requestedURI);
                    LocalService service = (LocalService) resource.getModel();

                    ServiceDescriptorBinder serviceDescriptorBinder =
                            getUpnpService().getConfiguration().getServiceDescriptorBinderUDA10();
                    String serviceDescriptor = serviceDescriptorBinder.generate(service);
                    return new StreamResponseMessage(
                            serviceDescriptor,
                            new ContentTypeHeader(ContentTypeHeader.DEFAULT_CONTENT_TYPE)
                    );

                case ICON:

                    log.fine("Found local icon matching relative request URI: " + requestedURI);
                    Icon icon = (Icon) resource.getModel();
                    return new StreamResponseMessage(icon.getData(), icon.getMimeType());

                default:
                    log.fine("Ignoring GET for found local resource: " + requestedURI);
                    return null;
            }
        } catch (DescriptorBindingException ex) {
            log.warning("Error generating requested device/service descriptor: " + ex.toString());
            log.log(Level.WARNING, "Exception root cause: ", Exceptions.unwrap(ex));
            return new StreamResponseMessage(UpnpResponse.Status.INTERNAL_SERVER_ERROR);
        }
    }

}
